#include <windows.h>
#include <vcclr.h>
#include <sstream>
#using <System.dll>
#include "libsass\sass_interface.h"
#include "libsass\document.hpp"
#include "libsass\eval_apply.hpp"
#include "libsass\error.hpp"

extern "C" { static char* process_document(Sass::Document& doc, int style); }

#pragma managed

namespace XWeb
{
  namespace Web
  {
    namespace Sass
    {
      using namespace System;
      using namespace System::Runtime::InteropServices;
      using namespace System::Reflection;

      [assembly: AssemblyTitle("LibSass.Net")];
      [assembly: AssemblyDescription("Wrapper for libsass library")];
      [assembly: AssemblyConfiguration("")];
      [assembly: AssemblyCompany("DigitSoft")];
      [assembly: AssemblyProduct("XWeb.Web.Sass.Core")];
      [assembly: AssemblyCopyright("Copyright   2012")];
      [assembly: AssemblyTrademark("")];
      [assembly: AssemblyCulture("")];

      [assembly: AssemblyVersion("0.2.201.0")];
      [assembly: AssemblyFileVersion("0.2.201.0")];

      [System::SerializableAttribute]
      public ref class SassCompileException: Exception
      {
      public:
        SassCompileException() {}
        SassCompileException(String^ message): Exception(message) {}
        SassCompileException(String^ message, Exception^ innerException): Exception(message, innerException) {}
      public protected:
        SassCompileException(System::Runtime::Serialization::SerializationInfo^ info, System::Runtime::Serialization::StreamingContext context): Exception(info, context) {}
      };

      public enum struct OutputStyle: int
      {
        Nested = 0,
        Expanded = 1,
        Compact = 2,
        Compressed = 3
      };

      public interface class ISourceProvider
      {
        property String^ Path
        {
          String^ get();
        }
        property String^ Content
        {
          String^ get();
        }
        ISourceProvider^ Resolve(String^ path);
      };

      std::string ConvertToString(String^ source)
      {
        auto encodedBytes = System::Text::Encoding::UTF8->GetBytes(source);
        pin_ptr<Byte> pinnedBytes = &encodedBytes[0];
        std::string result(reinterpret_cast<char*>(pinnedBytes), encodedBytes->Length);
        return result;
      }

      String^ ConvertFromCstr(const char* chars)
      {
        auto len = std::strlen(chars) * sizeof(*chars);
        auto bytes = gcnew array<Byte>(len);
        {
          pin_ptr<Byte> pinnedBytes = &bytes[0];
          std::memcpy(pinnedBytes, chars, len);
        }
        return System::Text::Encoding::UTF8->GetString(bytes);
      }

      String^ ConvertFromString(std::string& str)
      {
        auto len = str.length();
        auto bytes = gcnew array<Byte>(len * sizeof(std::string::value_type));
        {
          pin_ptr<Byte> pinnedBytes = &bytes[0];
          str._Copy_s(reinterpret_cast<char*>(pinnedBytes), len, len);
        }
        return System::Text::Encoding::UTF8->GetString(bytes);
      }

      class ManagedSourceProvider: public ::Sass::SourceProvider
      {
        gcroot<ISourceProvider^> source;
      public:
        ManagedSourceProvider(ISourceProvider^ source)
          :source(source)
        {
        }

        virtual std::string get_path()
        {
          auto provider = source;
          auto path = provider->Path;
          if (path == nullptr)
            return 0;
          return ConvertToString(path);
        }

        virtual std::string get_content()
        {
          auto provider = source;
          auto content = provider->Content;
          if (content == nullptr)
            return 0;
          return ConvertToString(content);
        }

        virtual std::shared_ptr<SourceProvider> resolve_path(std::string path)
        {
          auto provider = source;
          auto new_source = provider->Resolve(ConvertFromString(path));
          if (new_source == nullptr)
            return 0;
          return std::shared_ptr<SourceProvider>(new ManagedSourceProvider(new_source));
        }
      };

      public ref class SassCompiler abstract sealed
      {
      public:
        static String^ CompileString(String^ input) { return CompileString(input, OutputStyle::Nested, nullptr); }
        static String^ CompileString(String^ input, OutputStyle style) { return CompileString(input, style, nullptr); }
        static String^ CompileString(String^ input, String^ includePaths) { return CompileString(input, OutputStyle::Nested, includePaths); }
        static String^ CompileString(String^ input, OutputStyle style, String^ includePaths)
        {
          if (!input)
            throw gcnew ArgumentNullException("input");

          auto inputPtr = IntPtr::Zero, includePathsPtr = IntPtr::Zero;
          auto context = sass_new_context();
          try
          {
            auto inputStr = ConvertToString(input);
            context->source_string = inputStr.c_str();

            context->options.output_style = safe_cast<int>(style);

            std::string includePathsStr;
            if (includePaths)
            {
              includePathsStr = ConvertToString(includePaths);
              context->options.include_paths = includePathsStr.c_str();
            }

            sass_compile(context);

            if (context->error_status)
              throw gcnew SassCompileException(ConvertFromCstr(context->error_message));

            return ConvertFromCstr(context->output_string);
          }
          finally
          {
            sass_free_context(context);
          }
        }

        static String^ CompileFile(String^ inputFilename) { return CompileFile(inputFilename, OutputStyle::Nested, nullptr); }
        static String^ CompileFile(String^ inputFilename, OutputStyle style) { return CompileFile(inputFilename, style, nullptr); }
        static String^ CompileFile(String^ inputFilename, String^ includePaths) { return CompileFile(inputFilename, OutputStyle::Nested, includePaths); }
        static String^ CompileFile(String^ inputFilename, OutputStyle style, String^ includePaths)
        {
          if (!inputFilename)
            throw gcnew ArgumentNullException("input");

          auto context = sass_new_file_context();
          try
          {
            auto inputFilenameStr = ConvertToString(inputFilename);
            context->input_path = inputFilenameStr.c_str();

            context->options.output_style = safe_cast<int>(style);

            std::string includePathsStr;
            if (includePaths)
            {
              includePathsStr = ConvertToString(includePaths);
              context->options.include_paths = includePathsStr.c_str();
            }

            sass_compile_file(context);

            if (context->error_status)
              throw gcnew SassCompileException(ConvertFromCstr(context->error_message));

            return ConvertFromCstr(context->output_string);
          }
          finally
          {
            sass_free_file_context(context);
          }
        }

        static String^ CompileSource(ISourceProvider^ source, OutputStyle style)
        {
          using namespace ::Sass;

          if (!source)
            throw gcnew ArgumentNullException("source");

          Context ctx("");
          std::shared_ptr<SourceProvider> source_provider(new ManagedSourceProvider(source));
          Document doc(Document::make_from_source_provider(ctx, source_provider));

          try
          {
            doc.parse_scss();
            eval(doc.root,
              doc.context.new_Node(Node::none, doc.path, doc.line, 0),
              doc.context.global_env,
              doc.context.function_env,
              doc.context.new_Node,
              doc.context);
            extend_selectors(doc.context.pending_extensions, doc.context.extensions, doc.context.new_Node);

            string output(doc.emit_css(static_cast<Document::CSS_Style>(safe_cast<int>(style))));

            return ConvertFromString(output);
          }
          catch (Error& e)
          {
            std::stringstream msg_stream;
            msg_stream << "ERROR -- " << e.path << ", line " << e.line << ": " << e.message << endl;
            throw gcnew SassCompileException(ConvertFromString(msg_stream.str()));
          }
          catch(bad_alloc& ba)
          {
            std::stringstream msg_stream;
            msg_stream << "ERROR -- unable to allocate memory: " << ba.what() << endl;
            throw gcnew SassCompileException(ConvertFromString(msg_stream.str()));
          }
        }
      };
    }
  }
}